// Create the canvas
var canvas = document.createElement("canvas");
var ctx = canvas.getContext("2d");
canvas.width = 800;
canvas.height = 400;
document.body.appendChild(canvas);

var cx=canvas.width/2;
var cy=canvas.height;
ctx.translate(cx,cy);

var boardWidth = canvas.width + 320;

var screenScale = function () {
	return canvas.width * 1.0 / boardWidth * 1.0;
}

var loadedImageDict = {};

var loadImage = function (imageName) {
	var bgImage = new Image();
	bgImage.onload = function () {
		loadedImageDict[imageName] = bgImage;
	};
	bgImage.src = "images/"+imageName+".png";
};

var getImageIfReady = function (imageName) {
	return loadedImageDict[imageName];
}

var prepareAllImages = function () {
	
};

// Game objects
var LAUNCH_STATE_GAIN_POWER = 1;
var LAUNCH_STATE_GAIN_ANGLE = 2;
var LAUNCH_STATE_FLYING = 3;
var LAUNCH_STATE_GAMEOVER = 4;
var LAUNCH_STATE_WAITINGFORSTART = 4;
var player1 = {};
var player2 = {};
var stars = [];
var moon = [];
var particles = [];
var timeElapsed = 0;

var gravityY = 10;
var maxMarkers = 50;
var starMinLifetime = 3;
var starMaxLifetime = 5;
var shootTheMoonAlpha = 0;
var gameStartTimer = 3;

// Handle keyboard control
var P1_FIRE = 81;
var P2_FIRE = 80;
var keysDown = {};

// TODO: note that if you task out the key will never drop
addEventListener("keydown", function (e) {
	keysDown[e.keyCode] = true;
}, false);

addEventListener("keyup", function (e) {
	delete keysDown[e.keyCode];
}, false);

var spawnStar = function () {
	newStar = {};

 	// TODO: offset for sprite width? i'm lazy
 	minimumHeight = 48 * 4;

	newStar.x = (canvas.width - boardWidth) / 2 + (Math.random() * boardWidth) - 24; // don't ask why this formula works
	newStar.y = 600 - (Math.random() * (600 - minimumHeight) / screenScale()) - 48 - minimumHeight;
	loadImage("star");
	newStar.sprite = "star";
	newStar.lifetime = starMinLifetime + (Math.random() * (starMaxLifetime - starMinLifetime));
	newStar.collided = false;

	stars.push(newStar);
}

var spawnMoon = function () {
	moon = {};

	moon.x = canvas.width / 2 - 192 / 2; // TODO: sprite width
	moon.y = -1500;
	loadImage("moon");
	loadImage("moon_p1");
	loadImage("moon_p2");
	moon.sprite = "moon";
	moon.collided = false;
}

var spawnParticlesAtPoint = function (x, y, color) {
	var particlesC = 10 + Math.random() * 20;
	for (i = 0; i < particlesC; i++) {
		newParticle = {};

		newParticle.x = x;
		newParticle.y = y;
		newParticle.color = color;
		newParticle.velocityX = -5 + Math.random() * 10;
		newParticle.velocityY = -5 + Math.random() * 10;
		newParticle.lifetime = 1;
		newParticle.sprite = "star"; // TODO: lazy

		particles.push(newParticle);
	}
}

var spawnShootMoonParticles = function (sprite) {
	for (i = 0; i < 19; i++) {
		newParticle = {};

		newParticle.x = moon.x;
		newParticle.y = moon.y;
		newParticle.velocityX = 5 * Math.cos(Math.PI / 180 * i * 20);
		newParticle.velocityY = 5 * Math.sin(Math.PI / 180 * i * 20);
		newParticle.lifetime = 1;
		newParticle.sprite = sprite;

		particles.push(newParticle);
	}
}

var initPlayer = function (player, pSprite, pKey, aimingToRight) {
	player.powerAccel = 8;
	player.angleAccel = 3;
	player.maxPower = 800;
	player.maxAngle = 180;
	player.aimingRight = aimingToRight; 
	player.practiceShots = 3;

	player.score = 0;

	loadImage(pSprite);
	player.sprite = pSprite;
	player.fireKey = pKey;

	respawnPlayer(player);
};

var offsetPointForMarker = function (index) {
	base = canvas.width / 2 - 32;
	base += 32 * index;
	return base;
};

var respawnPlayer = function (player) {
	playerCenterOffset = offsetPointForMarker(player.score);
	player.x = canvas.width / 2 - playerCenterOffset;
	if (!player.aimingRight) {
		player.x = canvas.width / 2 + playerCenterOffset; // TODO: "sprite width" instead of 32
	}

	player.y = (canvas.height - 32 - 16); // TODO: screenscale to tweak

	player.launchPowerRadians = -1 * Math.PI / 2; 
	player.launchPower = 200;
	player.launchAngle = 0;
	player.launchState = LAUNCH_STATE_GAIN_POWER;
};

var updatePlayer = function (player, modifier) {
	if (player.launchState == LAUNCH_STATE_GAIN_POWER) {
		if (player.fireKey in keysDown) { 
			player.launchState = LAUNCH_STATE_GAIN_ANGLE;
		} else {
			player.launchPowerRadians += Math.PI / 45; //  TODO: power accel is unused
			player.launchPower = player.maxPower * (0.5 + 0.5 * Math.sin(player.launchPowerRadians));
		}
	} else if (player.launchState == LAUNCH_STATE_GAIN_ANGLE) {
		if (player.fireKey in keysDown && player.launchAngle < player.maxAngle) { 
			player.launchAngle += player.angleAccel;
		} else {
			player.launchState = LAUNCH_STATE_FLYING;

			// calculate initial velocity
			angleInRadians = Math.PI * 2 / 360 * player.launchAngle;
			player.velocityX = player.launchPower * Math.cos(angleInRadians);
			if (!player.aimingRight) {
				player.velocityX *= -1;
			}
			player.velocityY = player.launchPower * Math.sin(angleInRadians);
		}
	} else if (player.launchState == LAUNCH_STATE_FLYING) {
		player.x += player.velocityX * modifier;
		player.y -= player.velocityY * modifier;

		player.velocityY -= gravityY;

		if (player.y > canvas.height) { // off screen, reset player
			player.maxPower = 800 + player.score * 18;
			player.powerAccel = player.maxPower / 100.0;

			// TODO: practice!
			if (player.practiceShots > 0) {
				player.practiceShots -= 1;
				if (player.practiceShots != 0) { // wait for both players to ready
					respawnPlayer(player);
				} else {
					player.launchState = LAUNCH_STATE_WAITINGFORSTART;
				}
			} else {
				respawnPlayer(player);
			}

			var maxScore = Math.max(player1.score, player2.score);
			if (maxScore < 5) {
				maxScore = 5;
			}
			if (player.score == maxScore) {
				boardWidth = canvas.width + 64 * maxScore;
			}
		}
	} else if (player.launchState == LAUNCH_STATE_GAMEOVER) {
		player.x += player.velocityX * modifier;
		player.y -= player.velocityY * modifier;

		player.velocityY -= gravityY;

		if (player.y > canvas.height * 2) { 
			player.velocityY = 0;
		}
	}
};

var updateMoon = function(star, modifier) {
	star.lifetime -= modifier;

	// TODO: this kinda sux
	if (star.x < player1.x && player1.x < star.x + 192 && star.y < player1.y && player1.y < star.y + 192 && !star.collided) {
		star.collided = true;
		star.sprite = "moon_p1";
		player1.launchState = LAUNCH_STATE_GAMEOVER;
		player2.launchState = LAUNCH_STATE_GAMEOVER;
		spawnShootMoonParticles("moon_p1");
	}

	if (star.x < player2.x && player2.x < star.x + 192 && star.y < player2.y && player2.y < star.y + 192 && !star.collided) {
		star.collided = true;
		star.sprite = "moon_p2";
		player1.launchState = LAUNCH_STATE_GAMEOVER;
		player2.launchState = LAUNCH_STATE_GAMEOVER;
		spawnShootMoonParticles("moon_p2");
	}
}


var updateStar = function(star, modifier) {
	star.lifetime -= modifier;

	// TODO: this kinda sux
	if (star.x < player1.x && player1.x < star.x + 48 && star.y < player1.y && player1.y < star.y + 48) {
		player1.score += 1;
		if (player1.score > maxMarkers) {
			player1.score = maxMarkers;
		}
		star.collided = true;
	}

	if (star.x < player2.x && player2.x < star.x + 48 && star.y < player2.y && player2.y < star.y + 48) {
		player2.score += 1;
		if (player2.score > maxMarkers) {
			player2.score = maxMarkers;
		}
		star.collided = true;
	}
}

// Update game objects
var update = function (modifier) {
	if (player1.practiceShots == 0 && player2.practiceShots == 0) {
		if (!gameStarted) {
			gameStartTimer -= modifier;
			if (gameStartTimer <= 0) {
				startGame();
			}
		} else {
			if (player1.launchState != LAUNCH_STATE_GAMEOVER) {
				timeElapsed += modifier;
			} else {
				if (shootTheMoonAlpha < 1) {
					shootTheMoonAlpha += modifier / 3;
				}
			}
		}
	}

	updatePlayer(player1, modifier);
	updatePlayer(player2, modifier);

	// update particles
	var i = 0;
	while (i < particles.length && particles.length > 0) {
		particles[i].lifetime -= modifier;
		particles[i].x += particles[i].velocityX;
		particles[i].y += particles[i].velocityY;
		if (particles[i].sprite != "moon") {
			particles[i].velocityY += 0.1;
		}
	
		if (particles[i].lifetime <= 0) { // despawn this
			particles.splice(i, 1);

			i -= 1;
		}

		i += 1;
	}

	if (moon)
	{
		updateMoon(moon);
	}

	// update stars
	var i = 0;
	var starsToSpawn = 0;
	while (i < stars.length && stars.length > 0) {
		updateStar(stars[i], modifier);

		if (stars[i].lifetime <= 0 || stars[i].collided) { // despawn this star
			starsToSpawn += 1;
			if (stars[i].collided) { 
				starsToSpawn += 1;
				spawnParticlesAtPoint(stars[i].x, stars[i].y)
			}
			
			stars.splice(i, 1);

			i -= 1;
		}

		i += 1;
	}

	i = 0;
	while (i < starsToSpawn) {
		spawnStar();
		i += 1;
	}
};

var renderStar = function (star) {
	if (getImageIfReady(star.sprite)) {
		ctx.globalAlpha = star.lifetime / starMinLifetime;
		ctx.drawImage(getImageIfReady(star.sprite), star.x - cx, star.y - cy);
		ctx.globalAlpha = 1;
	}
}

var renderMarkers = function () {
	var i = 0;
	while (i <= maxMarkers) {
		playerCenterOffset = offsetPointForMarker(i);

		var x = canvas.width / 2 - playerCenterOffset;
		var x2 = canvas.width / 2 + playerCenterOffset; // TODO: "sprite width" instead of 32

		if (i == maxMarkers) {
			ctx.strokeStyle = "#FFFF00";
		} else if (i == player1.score) {
			ctx.strokeStyle = "#00FF00";
		} else if (i % 5 == 0) {
			ctx.strokeStyle = "#FF0000";
		} else {
			ctx.strokeStyle = "#000000";
		}
		
		ctx.lineWidth = 3 / screenScale();
		ctx.beginPath();			
		ctx.moveTo(x - cx,(canvas.height - 32 - 16) - cy);
		ctx.lineTo(x - cx,canvas.height - cy);
		ctx.stroke();

		if (i == maxMarkers) {
			ctx.strokeStyle = "#FFFF00";
		} else if (i == player2.score) {
			ctx.strokeStyle = "#00FF00";
		} else if (i % 5 == 0) {
			ctx.strokeStyle = "#FF0000";
		} else {
			ctx.strokeStyle = "#000000";
		}

		ctx.beginPath();			
		ctx.moveTo(x2 - cx,(canvas.height - 32 - 16) - cy);
		ctx.lineTo(x2 - cx,canvas.height - cy);
		ctx.stroke();

		i += 1;
	}
};

var renderPlayer = function (player) {
	if (getImageIfReady(player.sprite)) {
		ctx.drawImage(getImageIfReady(player.sprite), player.x - cx - 16, player.y - cy - 16);
	}

	//ctx.fillText("Power: " + player1.launchPower + " / Angle: " + player1.launchAngle, 16, 16);
	playerSpriteMod = 16;
	if (player.aimingRight) {
		playerSpriteMod = -16;
	}

	if (player.launchState < LAUNCH_STATE_FLYING) {
		// TODO: make this less ugly
		ctx.strokeStyle = "#000000";
		ctx.lineWidth = 5 / screenScale();
		ctx.beginPath();			
		arrowLength = 50.0 * (player.maxPower / 800) / screenScale();
		angleInRadians = Math.PI * 2 / 360 * player.launchAngle;
		arrowDirection = 1;
		if (!player.aimingRight) {
			arrowDirection *= -1;
		}
		angleLineX = player.x + arrowLength * Math.cos(angleInRadians) * arrowDirection;
		angleLineY = player.y - arrowLength * Math.sin(angleInRadians);
		ctx.moveTo(player.x - cx - playerSpriteMod,player.y - cy);
		ctx.lineTo(angleLineX - cx - playerSpriteMod, angleLineY - cy);
		ctx.stroke();

		ctx.strokeStyle = "#FF5555";
		ctx.beginPath();			
		arrowLength = 50.0 * (player.maxPower / 800) * player.launchPower / player.maxPower / screenScale();
		angleInRadians = Math.PI * 2 / 360 * player.launchAngle;
		arrowDirection = 1;
		if (!player.aimingRight) {
			arrowDirection *= -1;
		}
		angleLineX = player.x + arrowLength * Math.cos(angleInRadians) * arrowDirection;
		angleLineY = player.y - arrowLength * Math.sin(angleInRadians);
		ctx.moveTo(player.x - cx - playerSpriteMod,player.y - cy);
		ctx.lineTo(angleLineX - cx - playerSpriteMod, angleLineY - cy);
		ctx.stroke();
	}
}

// Draw everything
var currentScale = 1.0;
var maxBoardHeight = 6000;
var maxBoardWidth = 8000;

var render = function () {
	scrScale = screenScale();

	if (scrScale != currentScale) {
		ctx.scale(screenScale() / currentScale, screenScale() / currentScale);
		currentScale = screenScale();
	}
	
	var BGgrd=ctx.createLinearGradient(0 - cx,0 - cy / screenScale(),0 - cx,(600 - cy) / screenScale());
	bg1 = Math.floor(50 - 30 * (Math.max(player1.score, player2.score) * 1.0 / maxMarkers));
	bg2 = Math.floor(150 - 100 * (Math.max(player1.score, player2.score) * 1.0 / maxMarkers));
	BGgrd.addColorStop(0,"rgb("+bg1+", "+bg1+", "+bg2+")");
	BGgrd.addColorStop(1,"rgb("+Math.floor(bg1*8.0/5)+", "+Math.floor(bg1*8.0/5)+", "+Math.floor(bg2*185.0/150)+")");

	ctx.fillStyle=BGgrd;
	ctx.fillRect(0 - cx / screenScale(),0 - cy / screenScale(),boardWidth,600 / screenScale());

	// Score
	renderMarkers();

	// draw timer
	ctx.fillStyle = "rgb(250, 250, 250)";
	ctx.textAlign = "center";
	ctx.textBaseline = "top";
	ctx.fillStyle="#FFFFFF";
	ctx.font = "64px Helvetica";
	var timerString = timeElapsed.toFixed(3);
	inTutorial = false;
	if (player1.practiceShots > 0 || player2.practiceShots > 0 || gameStartTimer > 0) {
		timerString = "Practice"
		if (gameStartTimer < 3) {
			timerString = "Start in " + Math.ceil(gameStartTimer.toFixed(3));
			ellipseCount = Math.ceil((gameStartTimer % 1) / 0.2);
			for (i = 0; i < ellipseCount; i++) {
				timerString += ".";
			}
		}
		inTutorial = true;
	}
	ctx.fillText(timerString, (canvas.width / 2 - cx) / screenScale(), -160 - cy);

	ctx.font = "32px Helvetica";
	if (inTutorial) {
		if (player1.practiceShots == 3) {
			p1String = "Hold + release Q";
		} else if (player1.practiceShots == 2) {
			p1String = "Long hold = big angle";
		} else if (player1.practiceShots == 1) {
			p1String = "Full meter = full power";
		} else {
			p1String = "Shoot the moon!";
		}

		if (player2.practiceShots == 3) {
			p2String = "Hold + release P";
		} else if (player2.practiceShots == 2) {
			p2String = "Long hold = big angle";
		} else if (player2.practiceShots == 1) {
			p2String = "Full meter = full power";
		} else {
			p2String = "Shoot the moon!";
		}

		ctx.globalAlpha = gameStartTimer / 3;
		ctx.fillText(p1String, (canvas.width * 0.2 - 24 - cx) / screenScale(), 272 - cy);
		ctx.fillText(p2String, (canvas.width * 0.8 + 24 - cx) / screenScale(), 272 - cy);
		ctx.globalAlpha = 1;
	}
	

	var i = 0;
	while (i < stars.length) {
		renderStar(stars[i]);
		i += 1;
	}

	renderPlayer(player1);
	renderPlayer(player2);

	// render particles
	var i = 0;
	while (i < particles.length) {
		renderStar(particles[i]);
		i += 1;
	}

	if (moon) {
		renderStar(moon);
	}

	if (shootTheMoonAlpha > 0) {
		ctx.font = "256px Helvetica";
		ctx.globalAlpha = shootTheMoonAlpha;
		ctx.fillText("Shoot the Moon", (canvas.width / 2 - cx) / screenScale(), -800 - cy);
		ctx.globalAlpha = 1;
	}
};

var init = function () {
	prepareAllImages();

	timeElapsed = 0;
	gameStarted = false;

	spawnMoon();
	shootTheMoonAlpha = -1;
	gameStartTimer = 3;
	
	initPlayer(player1, "p1", P1_FIRE, true);
	initPlayer(player2, "p2", P2_FIRE, false);
	main();
}

var startGame = function() {
	gameStarted = true;

	respawnPlayer(player1);
	respawnPlayer(player2);

	for (i = 0; i < 5; i++) {
		spawnStar();
	}
}

// The main game loop
var main = function () {
	var now = Date.now();
	var delta = now - then;

	update(delta / 1000);
	render();

	then = now;

	// Request to do this again ASAP
	requestAnimationFrame(main);
};

// Cross-browser support for requestAnimationFrame
var w = window;
requestAnimationFrame = w.requestAnimationFrame || w.webkitRequestAnimationFrame || w.msRequestAnimationFrame || w.mozRequestAnimationFrame;

// Let's play this game!
var then = Date.now();
init();
